################################################################################
############################# INNER FUNCTIONS ##################################
################################################################################
#' snow.season.length
#'
#' \code{snow.season.length()} is an internal function called by
#' [plt.snow.season.length()]), calling a C++ function returning the snow
#' season length foe each pixels
#'
#' @param data The dataset (a vectdata object, see [vectorize.list()])
#' @param ncores Number of cores to use for parallel computation, using ncores =
#' 1 defaults to serial computing
#' @param swe.threshold The threshold value to define snow day in order to
#' compute the length of the snow season
#' @param ma.value Number of day to apply for moving average function on SWE
#' before computing the length of the snow seasons
#'
#' @return A list containing for each pixel a vector containing the length of
#' the snow season for each individual year
#'
#' @author Adrien Michel, 2023 (MeteoSwiss & WSL/SLF)
snow.season.length <- function(data, ncores = 1, swe.threshold = 0.05,
                               ma.value = 7){
  #Compute indices list for each year
  years <- lubridate::year(data$time.date)
  years <- ifelse(lubridate::month(data$time.date) %in% c(9, 10, 11, 12),
                  years+1, years)
  indices=list()
  for(yr in unique(years)){
    indices[[paste(yr)]]=which(years==yr)
  }

  snow.season <- snowSeasonLengthCpp(data$data, indices, ncores, swe.threshold,
                                     ma.value)

  return(snow.season)
}


################################################################################
############################# SINGLE MODEL RUN #################################
################################################################################
#' plt.snow.season.length
#'
#' \code{plt.snow.season.length} produces computes the snow season length for
#' the three datasets passed as arguments. It returns the average and absolute 
#' average error in snow season length for the model and corrected dataset
#' compared to the training one. 
#'
#' @param data.training The training dataset (a vectdata object, see
#' [vectorize.list()])
#' @param data.model The model dataset (a vectdata object, see
#' [vectorize.list()])
#' @param data.corrected The corrected dataset (a vectdata object, see
#' [vectorize.list()])
#' @param ncores Number of cores to use for parallel computation, using ncores =
#' 1 defaults to serial computing
#' @param out.plot.dir Path in which the RDS files should be saved. Three files are
#' created: snow_season_training, snow_season_model, snow_season_corrected
#' @param swe.threshold The threshold value to define snow day in order to
#' compute the length of the snow season
#' @param ma.value Number of day to apply for moving average function on SWE
#' before computing the length of the snow seasons.
#'
#' @return A list containing the mean error and mean absolute error in snow 
#' season length for the model and training data. It also returns three 
#' [vectorize.list()] objects containing teh average snow season length 
#' for each pixel in the \code{snow.season} list entry. 
#'
#' @author Adrien Michel, 2023 (MeteoSwiss & WSL/SLF)
#' @export
plt.snow.season.length <- function(data.training, data.model,
                                   data.corrected, ncores,
                                   swe.threshold, ma.value) {

  # Compute snow season length for each pixels
  snow.season.training <- snow.season.length(data.training, ncores,
                                                 swe.threshold, ma.value)
  snow.season.model <- snow.season.length(data.model, ncores,
                                                   swe.threshold, ma.value)
  snow.season.corrected <- snow.season.length(data.corrected, ncores,
                                                    swe.threshold, ma.value)

  # Plot relative error calibration
  snow.season.model.rel <- (snow.season.model-snow.season.training)/snow.season.training
  snow.season.corrected.rel <- (snow.season.corrected-snow.season.training)/snow.season.training
  snow.season.model.rel[snow.season.model.rel>1] <- 1
  snow.season.corrected.rel[snow.season.corrected.rel>1] <- 1
  snow.season.model.rel[is.na(snow.season.model.rel)] <- 0
  snow.season.corrected.rel[is.na(snow.season.corrected.rel)] <- 0

  # Prepare output data
  out=list()
  out$model$mean.error <- mean(snow.season.model.rel)
  out$model$mae <- mean(abs(snow.season.model.rel))
  out$corrected$mean.error <- mean(snow.season.corrected.rel)
  out$corrected$mae <- mean(abs(snow.season.corrected.rel))

  # Add data to produce map plots from output
  data.training$snow.season <- matrix(rowMeans(snow.season.training))
  data.training$data <- NULL
  out$data.training <- data.training
  data.model$snow.season <- matrix(rowMeans(snow.season.model))
  data.model$data <- NULL
  out$data.model <- data.model
  data.corrected$snow.season <- matrix(rowMeans(snow.season.corrected))
  data.corrected$data <- NULL
  out$data.corrected <- data.corrected

  return(out)
}

#' plt.snow.season.maps
#'
#' \code{plt.snow.season.maps} produces maps of average snow season length for
#' each pixels, returned by [plt.snow.season.length()]. It also plots,
#' for each pixels, the relative error of the average snoe season length in
#' the model and corrected dataset.
#'
#' @param snow.season.training The snow average snow season length (as a vectdata 
#' object,  see [vectorize.list()]) for the training dataset, as returned by 
#' \code{plt.snow.season.length}
#' @param snow.season.model The snow average snow season length (as a vectdata 
#' object,  see [vectorize.list()]) for the model dataset, as returned by 
#' \code{plt.snow.season.length}
#' @param snow.season.corrected The snow average snow season length (as a vectdata 
#' object,  see [vectorize.list()]) for the corrected dataset, as returned by 
#' \code{plt.snow.season.length}
#' @param out.plot.dir Directory where output plots should be saved
#'
#' @author Adrien Michel, 2023 (MeteoSwiss & WSL/SLF)
#' @export
plt.snow.season.maps <- function(snow.season.training, snow.season.model,
                                 snow.season.corrected, out.plot.dir) {

  pdf(paste0(out.plot.dir,"/snow_season_maps.pdf"),width=16,height=6)

  par(mar=c(3,4,2,7), mgp=c(1.7,0.5,0),mfrow=c(1,3))

  #Plot map of absolute values
  plot(snow.season.training, var="snow.season", zlim=c(0,366),
       main="training", legend.lab="Snow season length (days)",
       legend.cex=0.7, legend.line=2.7)
  plot(snow.season.model, var="snow.season", zlim=c(0,366),
       main="model", legend.lab="Snow season length (days)",
       legend.cex=0.7, legend.line=2.7)
  plot(snow.season.corrected, var="snow.season", zlim=c(0,366),
       main="Corrected", legend.lab="Snow season length (days)",
       legend.cex=0.7, legend.line=2.7)

  par(mar=c(3,4,2,8), mgp=c(1.7,0.5,0),mfrow=c(1,2))
  #Compute relative error
  snow.season.model$snow.season.rel <- (snow.season.model$snow.season -
                                                   snow.season.training$snow.season) /
                                                 snow.season.training$snow.season
  snow.season.corrected$snow.season.rel <- (snow.season.corrected$snow.season -
                                                    snow.season.training$snow.season) /
                                                  snow.season.training$snow.season
  snow.season.model$snow.season.rel[snow.season.model$snow.season.rel>1] <- 1
  snow.season.corrected$snow.season.rel[snow.season.corrected$snow.season.rel>1] <- 1

  #Plot relative error
  plot(snow.season.model, var="snow.season.rel", zlim=c(-1,1),
       main="Model", legend.lab="Error (%)",
       legend.cex=0.9, legend.line=2.7)
  plot(snow.season.corrected, var="snow.season.rel", zlim=c(-1,1),
       main="Corrected", legend.lab="Error (%)",
       legend.cex=0.9, legend.line=2.7)
  dev.off()

}

################################################################################
####################### CALIBRATION AND VALIDATION #############################
################################################################################

#' plt.snow.season.length.calib
#'
#' \code{plt.snow.season.length.calib} produces computes the snow season length 
#' for the three datasets passed as arguments. It returns the average and absolute 
#' average error in snow season length for the model and corrected dataset
#' compared to the training one. Computation are performed for both the 
#' calibration and validation periods.
#'
#' @param data.training.calib The training dataset over the calibration period 
#' (a vectdata object, see[vectorize.list()])
#' @param data.model.calib The model dataset over the calibration period 
#' (a vectdata object, see[vectorize.list()])
#' @param data.corrected.calib The corrected dataset over the calibration period 
#' (a vectdata object, see[vectorize.list()])
#'  @param data.training.valid The training dataset over the validation period 
#' (a vectdata object, see[vectorize.list()])
#' @param data.model.valid The model dataset over the validation period 
#' (a vectdata object, see[vectorize.list()])
#' @param data.corrected.valid The corrected dataset over the validation period 
#' (a vectdata object, see[vectorize.list()])
#' @param ncores Number of cores to use for parallel computation, using ncores =
#' 1 defaults to serial computing
#' @param swe.threshold The threshold value to define snow day in order to
#' compute the length of the snow season
#' @param ma.value Number of day to apply for moving average function on SWE
#' before computing the length of the snow seasons.
#'
#' @return A list containing the mean error and mean absolute error in snow 
#' season length for the model and training data. It also returns three 
#' [vectorize.list()] objects containing teh average snow season length 
#' for each pixel in the \code{snow.season} list entry. All datare are both 
#' for the calibration and validation periods.
#'
#' @author Adrien Michel, 2023 (MeteoSwiss & WSL/SLF)
#' @export
plt.snow.season.length.calib <- function(data.training.calib, data.model.calib,
                                         data.corrected.calib, data.training.valid,
                                         data.model.valid, data.corrected.valid,
                                         ncores, swe.threshold, ma.value) {

  # Compute snow season length for each pixels
  snow.season.training.calib <- snow.season.length(data.training.calib, ncores,
                                                 swe.threshold, ma.value)
  snow.season.model.calib <- snow.season.length(data.model.calib, ncores,
                                                   swe.threshold, ma.value)
  snow.season.corrected.calib <- snow.season.length(data.corrected.calib, ncores,
                                                    swe.threshold, ma.value)

  snow.season.training.valid <- snow.season.length(data.training.valid, ncores,
                                                 swe.threshold, ma.value)
  snow.season.model.valid <- snow.season.length(data.model.valid, ncores,
                                                   swe.threshold, ma.value)
  snow.season.corrected.valid <- snow.season.length(data.corrected.valid, ncores,
                                                    swe.threshold, ma.value)

  #Plot absolute values calibration
  calib.years <- colnames(snow.season.training.calib)
  data.length <- length(calib.years)

  #Plot relative error calibration
  snow.season.model.calib.rel <- (snow.season.model.calib-snow.season.training.calib)/snow.season.training.calib
  snow.season.corrected.calib.rel <- (snow.season.corrected.calib-snow.season.training.calib)/snow.season.training.calib
  snow.season.model.calib.rel[snow.season.model.calib.rel>1]<-1
  snow.season.corrected.calib.rel[snow.season.corrected.calib.rel>1]<-1
  snow.season.model.calib.rel[is.na(snow.season.model.calib.rel)]<-0
  snow.season.corrected.calib.rel[is.na(snow.season.corrected.calib.rel)]<-0

  #Plot absolute values validation
  valid.years <- colnames(snow.season.training.valid)
  data.length <- length(valid.years)
  snow.season.model.valid.rel <- (snow.season.model.valid-snow.season.training.valid)/snow.season.training.valid
  snow.season.corrected.valid.rel <- (snow.season.corrected.valid-snow.season.training.valid)/snow.season.training.valid
  snow.season.model.valid.rel[snow.season.model.valid.rel>1]<-1
  snow.season.corrected.valid.rel[snow.season.corrected.valid.rel>1]<-1
  snow.season.model.valid.rel[is.na(snow.season.model.valid.rel)]<-0
  snow.season.corrected.valid.rel[is.na(snow.season.corrected.valid.rel)]<-0

  #Prepare output data
  out=list()
  out$model$calib$mean.error <- mean(snow.season.model.calib.rel)
  out$model$calib$mae <- mean(abs(snow.season.model.calib.rel))
  out$corrected$calib$mean.error <- mean(snow.season.corrected.calib.rel)
  out$corrected$calib$mae <- mean(abs(snow.season.corrected.calib.rel))
  out$model$valid$mean.error <- mean(snow.season.model.valid.rel)
  out$model$valid$mae <- mean(abs(snow.season.model.valid.rel))
  out$corrected$valid$mean.error <- mean(snow.season.corrected.valid.rel)
  out$corrected$valid$mae <- mean(abs(snow.season.corrected.valid.rel))

  #Add data to produce map plots from output
  data.training.calib$snow.season <- matrix(rowMeans(snow.season.training.calib))
  data.training.calib$data <- NULL
  out$data.training.calib <- data.training.calib
  data.model.calib$snow.season <- matrix(rowMeans(snow.season.model.calib))
  data.model.calib$data <- NULL
  out$data.model.calib <- data.model.calib
  data.corrected.calib$snow.season <- matrix(rowMeans(snow.season.corrected.calib))
  data.corrected.calib$data <- NULL
  out$data.corrected.calib <- data.corrected.calib

  data.training.valid$snow.season <- matrix(rowMeans(snow.season.training.valid))
  data.training.valid$data <- NULL
  out$data.training.valid <- data.training.valid
  data.model.valid$snow.season <- matrix(rowMeans(snow.season.model.valid))
  data.model.valid$data <- NULL
  out$data.model.valid <- data.model.valid
  data.corrected.valid$snow.season <- matrix(rowMeans(snow.season.corrected.valid))
  data.corrected.valid$data <- NULL
  out$data.corrected.valid <- data.corrected.valid

  return(out)
}


#' plt.snow.season.maps
#'
#' \code{plt.snow.season.maps} produces maps of average snow season length for
#' each pixels, returned by [plt.snow.season.length.calib()]. It also plots,
#' for each pixels, the relative error of the average snow season length in
#' the model and corrected dataset, both for calibration and validation periods.
#'
#' @param snow.season.training.calib The snow average snow season length (as 
#' a vectdata  object,  see [vectorize.list()]) for the training dataset over
#' the calibration period, as returned by \code{plt.snow.season.length.calib}
#' @param snow.season.model.calib The snow average snow season length (as a 
#' vectdata  object,  see [vectorize.list()]) for the model dataset over
#' the calibration period, as returned by \code{plt.snow.season.length.calib}
#' @param snow.season.corrected.calib The snow average snow season length (as 
#' a vectdata object,  see [vectorize.list()]) for the corrected dataset over
#' the calibration period, as returned by \code{plt.snow.season.length.calib}
#' @param snow.season.training.valid The snow average snow season length (as 
#' a vectdata  object,  see [vectorize.list()]) for the training dataset over
#' the validation period, as returned by \code{plt.snow.season.length.valid}
#' @param snow.season.model.valid The snow average snow season length (as a 
#' vectdata  object,  see [vectorize.list()]) for the model dataset over
#' the validation period, as returned by \code{plt.snow.season.length.valid}
#' @param snow.season.corrected.valid The snow average snow season length (as 
#' a vectdata object,  see [vectorize.list()]) for the corrected dataset over
#' the validation period, as returned by \code{plt.snow.season.length.valid}
#' 
#' @param out.plot.dir Directory where output plots should be saved
#'
#' @author Adrien Michel, 2023 (MeteoSwiss & WSL/SLF)
#' @export
plt.snow.season.maps.calib <- function(snow.season.training.calib, snow.season.model.calib,
                                       snow.season.corrected.calib, snow.season.training.valid,
                                       snow.season.model.valid, snow.season.corrected.valid,
                                       out.plot.dir) {

  pdf(paste0(out.plot.dir,"/snow_season_maps.pdf"),width=12,height=8)

  par(mar=c(3,4,2,7), mgp=c(1.7,0.5,0),mfrow=c(2,3))

  #Plot map of absoulte values
  plot(snow.season.training.calib, var="snow.season", zlim=c(0,366),
       main="training calib", legend.lab="Snow season length (days)",
       legend.cex=0.7, legend.line=2.7)
  plot(snow.season.model.calib, var="snow.season", zlim=c(0,366),
       main="model calib", legend.lab="Snow season length (days)",
       legend.cex=0.7, legend.line=2.7)
  plot(snow.season.corrected.calib, var="snow.season", zlim=c(0,366),
       main="Corrected calib", legend.lab="Snow season length (days)",
       legend.cex=0.7, legend.line=2.7)
  plot(snow.season.training.valid, var="snow.season", zlim=c(0,366),
       main="training valid", legend.lab="Snow season length (days)",
       legend.cex=0.7, legend.line=2.7)
  plot(snow.season.model.valid, var="snow.season", zlim=c(0,366),
       main="model valid", legend.lab="Snow season length (days)",
       legend.cex=0.7, legend.line=2.7)
  plot(snow.season.corrected.valid, var="snow.season", zlim=c(0,366),
       main="Corrected valid", legend.lab="Snow season length (days)",
       legend.cex=0.7, legend.line=2.7)


  par(mar=c(3,4,2,8), mgp=c(1.7,0.5,0),mfrow=c(2,2))
  #Compute relative error
  snow.season.model.calib$snow.season.rel <- (snow.season.model.calib$snow.season -
                                                   snow.season.training.calib$snow.season) /
                                                  snow.season.training.calib$snow.season
  snow.season.corrected.calib$snow.season.rel <- (snow.season.corrected.calib$snow.season -
                                                    snow.season.training.calib$snow.season) /
                                                    snow.season.training.calib$snow.season
  snow.season.model.calib$snow.season.rel[snow.season.model.calib$snow.season.rel>1] <- 1
  snow.season.corrected.calib$snow.season.rel[snow.season.corrected.calib$snow.season.rel>1] <- 1

  snow.season.model.valid$snow.season.rel <- (snow.season.model.valid$snow.season -
                                                   snow.season.training.valid$snow.season) /
                                                   snow.season.training.valid$snow.season
  snow.season.corrected.valid$snow.season.rel <- (snow.season.corrected.valid$snow.season -
                                                    snow.season.training.valid$snow.season) /
                                                    snow.season.training.valid$snow.season
  snow.season.model.valid$snow.season.rel[snow.season.model.valid$snow.season.rel>1]<-1
  snow.season.corrected.valid$snow.season.rel[snow.season.corrected.valid$snow.season.rel>1]<-1

  #Plot relative error
  plot(snow.season.model.calib,var="snow.season.rel", zlim=c(-1,1),
       main="model calib", legend.lab="Error (%)",
       legend.cex=0.9, legend.line=2.7)
  plot(snow.season.corrected.calib,var="snow.season.rel", zlim=c(-1,1),
       main="Corrected calib", legend.lab="Error (%)",
       legend.cex=0.9, legend.line=2.7)
  plot(snow.season.model.valid,var="snow.season.rel", zlim=c(-1,1),
       main="model valid", legend.lab="Error (%)",
       legend.cex=0.9, legend.line=2.7)
  plot(snow.season.corrected.valid,var="snow.season.rel", zlim=c(-1,1),
       main="Corrected valid", legend.lab="Error (%)",
       legend.cex=0.9, legend.line=2.7)
  dev.off()

}
