#' apply.model
#'
#' \code{apply.model} applies a trained QM model (see
#' [compute.quantiles()]) to the dataset given as input.
#'
#' @param in.data.file Dataset on which the QM model should be applied. Dataset
#' should be a RDS file in a `vectData` format (see [vectorize.list()])
#' @param out.data.file Path of the file to write the corrected dataset. Dataset
#' will be a RDS file in a `vectData` format (see [vectorize.list()])
#' @param quantile.path Directory containing the directories with quantile distributions of the
#' training and target datasets (see [vectorize.list()])
#' @param ncores number of cores to be used
#' @param snow Is the quantile mapping applied to snow (will impact situations of 
#' null quantiles, see discussion in the scientific publication)
#' @param correct.first should the first time step be corrected (parameter used
#' for the operational implementation, users should keep it `TRUE` unless they
#' really know what they do). Has an impact only if `snow==true`
#'
#' @author Adrien Michel, 2023 (MeteoSwiss & WSL/SLF)
#' @export
apply.model <- function(in.data.file, out.data.file, quantile.path, ncores, snow = TRUE,
                        correct.first = TRUE)
{
  print("[I] Creating output directory")
  dir.create(dirname(out.data.file), showWarnings = FALSE)

  print("[I] Reading data")
  data <- readRDS(in.data.file)

  print("[I] Computing days of the year")
  doys=as.integer(strftime(data$time.date, format = "%j"))
  doys.unique=sort(unique(doys))
  `%do%` <- foreach::`%do%`
  doy.selection = foreach::foreach(doy=doys.unique) %do% {
    which(doys == doy)
  }
  names(doy.selection)=doys.unique

  print("[I] Correcting data")
  start.time <- Sys.time()
  corrected.swe <- applyQMCpp(data$data,
                              doyArray = doys,
                              xCoords=round(data$x.coords[data$x.indices]),
                              yCoords=round(data$y.coords[data$y.indices]),
                              trainingPath = paste0(quantile.path,"/training_quantiles"),
                              modelPath = paste0(quantile.path,"/model_quantiles"),
                              ncores = ncores,
                              snow = snow,
                              correctFirst = correct.first)
  data.corrected <- data
  data.corrected$data <- matrix(corrected.swe, ncol=dim(data$data)[2])

  print(paste("[I] Correction computed in",
              as.numeric(round(difftime(Sys.time(),start.time, units = "mins"),2)),
              "minutes"))

  saveRDS(data.corrected,out.data.file)

  print("[I] Cleaning memory")
  rm(list=ls())
  invisible(gc())
}

