#' plot.vectdata
#'
#' \code{plot.vectdata} is an overloaded of the `base` [plot()] function designed
#' to plot data from [vectorize.list()] objects. It plots a 2D map of the chosen
#' variable.
#'
#' @param vectdata Dataset containing the data to be plotted as a`vectData` object 
#' (see [vectorize.list()])
#' @param time Timestep to plot (in timesteo number)
#' @param main Main title of the plot to be printed
#' @param zlim A vector of limit to use in the color palette
#' @param var Variable to plot
#' @param ... Any other parameters to be passed to [fields::image.plot()] function
#'
#' @author Adrien Michel, 2023 (MeteoSwiss & WSL/SLF)
#' @export
plot.vectdata <- function(vectdata, time=NULL, main=NULL, zlim=NULL, var="data",...){
  if(!(var %in% names(vectdata))){
    stop(paste("Variable", var, "is not present int the vectdata provided"))
  }

  resolution = mean(vectdata$x.coords[-1] - vectdata$x.coords[-length(vectdata$x.coords)])
  x.shift = (min(vectdata$x.coords[vectdata$x.indices])-min(vectdata$x.coords))/resolution
  x.max = (max(vectdata$x.coords[vectdata$x.indices])-min(vectdata$x.coords))/resolution
  x.coords=vectdata$x.coords[(x.shift+1):(x.max+1)]

  y.shift = (min(vectdata$y.coords[vectdata$y.indices])-min(vectdata$y.coords))/resolution
  y.max = (max(vectdata$y.coords[vectdata$y.indices])-min(vectdata$y.coords))/resolution
  y.coords=vectdata$y.coords[(y.shift+1):(y.max+1)]

  data.plot <- matrix(NA, nrow=length(x.coords),
                      ncol=length(y.coords))


  # Check if variable to plot has a time dimension
  if(!is.null(dim(vectdata[[var]]))) {
    if(is.null(time)) {
      time <- 1
    }
    for (i in seq_along(vectdata$x.indices)) {
      data.plot[vectdata$x.indices[i] - x.shift, vectdata$y.indices[i] - y.shift] <-
        vectdata[[var]][i,time]
    }
    if(is.null(main)){
      main=paste(var,vectdata$time.date[time])
    }
  } else {
    for (i in seq_along(vectdata$y.indices)){
      time <- 1
      data.plot[vectdata$x.indices[i] - x.shift, vectdata$y.indices[i] - y.shift] <-
        vectdata[[var]][i,time]
    }
    if(is.null(main)) {
      main <- var
    }
  }

  if(is.null(zlim)){
    zlim <- c(min(data.plot,na.rm=TRUE), max(data.plot,na.rm=TRUE))
  }
  # For image.plot, zlim should be strictly increasing
  if(zlim[1]==zlim[2]){
    zlim[2] <- zlim[1]+0.1
  }
  fields::image.plot(x.coords, y.coords, data.plot, xlab="chx (m)",
                     ylab="chy (m)", main=main, zlim=zlim, ...)
}

#' plot.vectdatamask
#'
#' \code{plot.vectdatamask} plots a mask object as returned from
#' [mask.vectdata()].
#'
#' @param vectdatamask Mask to be plotted
#' @param main Main title to be printed on the plot
#' @param zlim A vector of limit to use in the color palette
#' @param ... Any other parameters to be passed to [fields::image.plot()] function
#'
#' @author Adrien Michel, 2023 (MeteoSwiss & WSL/SLF)
#' @export
plot.vectdatamask <- function(vectdatamask, main=NULL, zlim=NULL,...) {
  data.plot <- matrix(0, nrow=length(vectdatamask$x.coords),
                      ncol=length(vectdatamask$y.coords))
  for (i in seq_along(vectdatamask$y.indices)) {
    data.plot[vectdatamask$x.indices[i],vectdatamask$y.indices[i]] <- 1
  }
  if(is.null(main)) {
    main="mask"
  }
  if(is.null(zlim)) {
    zlim <- c(min(data.plot,na.rm=TRUE), max(data.plot,na.rm=TRUE))
  }
  # For image.plot, zlim should be strictly increasing
  if(zlim[1]==zlim[2]) {
    zlim[2] <- zlim[1]+0.1
  }
  fields::image.plot(vectdatamask$x.coords, vectdatamask$y.coords, data.plot,
                     xlab="chx (m)", ylab="chy (m)", main=main, zlim=zlim, ...)
}


#' plt.pixel
#'
#' \code{plt.pixel} plots a timeseries for the given pixel of a `vectData` object 
#' (see [vectorize.list()])
#'
#' @param data Dataset containing the data to be plotted (as a
#' [vectorize.list()] object)
#' @param x.coord Coordinates of the pixel along the X axis (in the base NetCDF
#' format)
#' @param y.coord Coordinates of the pixel along the Y axis (in the base NetCDF
#' format)
#' @param ... Any other parameters to be passed to [plot()] function
#'
#' @author Adrien Michel, 2023 (MeteoSwiss & WSL/SLF)
#' @export
plt.pixel <- function(data, x.coord, y.coord, ...) {
  pixel=which.min(sqrt((x.coord-data$x.coords[data$x.indices])^2 +
                         (y.coord-data$y.coords[data$y.indices])^2))

  plot(data$time.date,data$data[pixel,], type="l",
       main=paste("SWE", xcoord ,"-", ycoord, " pixel:",pixel),
       ylab="SWE (m)", xlab="", ...)
}